/* **************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * Condor Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-2003 Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.
 * Use of the CONDOR Software Program Source Code is authorized
 * solely under the terms of the Condor Public License (see LICENSE.TXT).
 * For more information contact:
 * CONDOR Team, Attention: Professor Miron Livny,
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685,
 * (608) 262-0856 or miron@cs.wisc.edu.
 * ***************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

package condor.classad;

import java.util.*;
import java.text.SimpleDateFormat;

/**
 * A library of builtin functions.
 * Each function is static and public, excepts an array of Expr as an argument
 * and returns an Expr as a result.
 * @see FuncCall
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.1
 */
public class Builtin {
    private static String VERSION = "$Id: Builtin.java,v 1.6 2003/10/10 17:13:31 solomon Exp $";


    /** The default constructor is made private to prevent anybody from trying
     * to create an instance of this class.
     */
    private Builtin() {
        throw new RuntimeException();
    } // Builtin()

    //============= string operations (see also string() below) ===============
        
    /** Converts each argument to a string and returns the concatenation
     * of the values.
     * @param args the arguments
     * @return the concatenation of the values.
     */
    public static Expr strcat(Expr[] args) {
        String res = "";
        for (int i = 0; i < args.length; i++) {
            res += getString(args[i]);
        }
        return Constant.getInstance(res);
    } // strcat(Expr[])

    /** Returns a substring of a string.
     * This is like the Perl version of substr.
     * Offset is the start of the substr, negative is from the end:
     * 0: first char of string
     * -1: last char of string
     * If length is omitted, it means the rest of the string.
     * If length is negative, that many characters are dropped.
     * @param args the arguments
     * @return the designated substring
     */
    public static Expr substr(Expr[] args) {
        if (args.length < 2 || args.length > 3) {
            return Constant.error("wrong number of args to substr");
        }
        if (args[0].type != Expr.STRING) {
            return Constant.error("substr: arg 1 must be string, not "
                                  + args[0].typeName());
        }
        String str = args[0].stringValue();

        // offset
        if (!(args[1].type == Expr.INTEGER)) {
            return Constant.error("substr: arg 2 must be int, not "
                                  + args[1].typeName());
        }
        int offset = args[1].intValue();
        // negative offset is from end of original string
        int strlen = str.length();
        if (offset < 0) {
            offset += strlen;
        }
        if (offset > strlen) {
            return Constant.getInstance("");
        }
        if (offset < 0) {
            offset = 0;
        }

        // length
        if (args.length == 3) {
            if (!(args[2].type == Expr.INTEGER)) {
                return Constant.error(
                    "substr: arg 3 must be int, not "
                    + args[2].typeName());
            }
            int len = args[2].intValue();
            // negative length means bytes to drop
            if (len < 0) {
                len += strlen - offset;
            }

            return
                Constant.getInstance(
                    len + offset > strlen
                        ? str.substring(offset)
                        : str.substring(offset,len+offset));
        } else return Constant.getInstance(str.substring(offset));
    } // substr(Expr[])

    /** Compares two strings.  The arguments are first converted to strings
     * as if by string().
     * @param s1 the first string.
     * @param s2 the second string.
     * @return an Integer expression with a value  that is less than zero if
     *         string(s1) &lt; string(s2), greater than zero if string(s1) &gt;
     *         string(s2), and equal to zero if string(s1) &eq; string(s2).
     */
    public static Expr strcmp(Expr s1, Expr s2) {
        return Constant.getInstance(getString(s1).compareTo(getString(s2)));
    } // strcmp(Expr, Expr)

    /** Compares two strings ignoring differences in case.  The arguments are
     * first converted to strings as if by string().
     * @param s1 the first string.
     * @param s2 the second string.
     * @return an Integer expression with a value  that is less than zero if
     *         string(s1) &lt; string(s2), greater than zero if string(s1) &gt;
     *         string(s2), and equal to zero if string(s1) &eq; string(s2).
     */
    public static Expr stricmp(Expr s1, Expr s2) {
        return Constant.getInstance(getString(s1)
                    .compareToIgnoreCase(getString(s2)));
    } // stricmp(Expr, Expr)

    /** Compares a string with a shell-style "glob" pattern.
     * The pat string may contain instances of * and/or ?.  The result is true
     * if pat matches str and false otherwise.  The pattern character ? matches
     * any character; * matches any sequence of zero or more characters.  A
     * backslash preceding any pattern character removes any special meaning
     * from that character and makes it match only itself.  For example \?
     * matches a question mark and \\ matches a single backslash.  Any other
     * character matches only itself.  Note that case is significant in the
     * comparison.
     * @param str the string to be compared against
     * @param pat the pattern to match
     * @return a boolean constant with value true if the pattern matches and
     * false otherwise.
    */
    public static Expr glob(Expr str, Expr pat) {
        if (!(str.type == Expr.STRING)) {
            return Constant.error("glob: arg 1 must be string, not "
                                  + str.typeName());
        }
        if (!(pat.type == Expr.STRING)) {
            return Constant.error("glob: arg 2 must be string, not "
                                  + pat.typeName());
        }
        return Constant.bool(glob(str.stringValue(),pat.stringValue(),false));
    } // glob(Expr, Expr)

    /** Compares a string with a shell-style "glob" pattern, ignoring case.
     * The pat string may contain instances of * and/or ?.  The result is true
     * if pat matches str and false otherwise.  The pattern character ? matches
     * any character; * matches any sequence of zero or more characters.  A
     * backslash preceding any pattern character removes any special meaning
     * from that character and makes it match only itself.  For example \?
     * matches a question mark and \\ matches a single backslash.  A lower-case
     * letter mathches the upper or lower case version of that letter.  Any
     * other character (including an upper-case letter) matches only itself.
     * @param str the string to be compared against
     * @param pat the pattern to match
     * @return a boolean constant with value true if the pattern matches and
     * false otherwise.
     */
    public static Expr iglob(Expr str, Expr pat) {
        if (!(str.type == Expr.STRING)) {
            return Constant.error("glob: arg 1 must be string, not "
                                  + str.typeName());
        }
        if (!(pat.type == Expr.STRING)) {
            return Constant.error("glob: arg 2 must be string, not "
                                  + pat.typeName());
        }
        return Constant.bool(glob(str.stringValue(),pat.stringValue(),true));
    } // iglob(Expr, Expr)

    //============= data conversion operations ===============================

    /** Convert to integer.  The arg is convered to an integer if possible.
     * Note: This method corresponds to int() in the ClassAd language, because
     * case is ignored in function names.  The name "int" cannot be used
     * because it is a reserved word in Java.
     * @param arg the value to be converted.
     * @return an integer constant corresponding to arg, or ERROR.
     */
    public static Expr Int(Expr arg) {
        switch (arg.type) {
        case Expr.INTEGER:
            return arg;
        case Expr.REAL:
            return Constant.getInstance((int)arg.realValue());
        case Expr.BOOLEAN:
            return Constant.getInstance(arg.isTrue() ? 1 : 0);
        case Expr.ABSOLUTE_TIME:
        case Expr.RELATIVE_TIME:
            long ms = ((Constant) arg).milliseconds();
            int secs = (int) (ms / 1000);
            return Constant.getInstance(secs);
        case Expr.STRING:
            String s = arg.stringValue();
            try {
                return Constant.getInstance((int) Constant.stringToDouble(s));
            } catch (NumberFormatException e) {
                return Constant.error("ill-formed integer " + s);
            }
        default:
            return Constant.error(
                "int(" + arg.typeName() + " " + arg.type + ")");
        }
    } // Int(Expr)

    /** Convert to real.  The arg is convered to a real if possible.
     * @param arg the value to be converted.
     * @return a real constant corresponding to arg, or ERROR.
     */
    public static Expr real(Expr arg) {
        if (arg.type == Expr.REAL) {
            return arg;
        }
        try {
            return Constant.getInstance(toReal(arg));
        } catch (NumberFormatException e) {
            return Constant.error("ill-formed real " + arg.stringValue());
        } catch (ClassCastException e) {
            return Constant.error("real(" + arg.typeName() + ")");
        }
    } // real(Expr)

    /** Converts the argument to a string.
     * @param s str the string to convert.
     * @return a string represenation of the Expr
     */
    public static Expr string(Expr s) {
        return s.type == Expr.STRING ? s : Constant.getInstance(getString(s));
    } // string(Expr)

    /** Convert to relative time.
     * @param arg the value to be converted.
     * @return a relative time or error constant.
     */
    public static Expr relTime(Expr arg) {
        if (arg.type == Expr.STRING) {
            return Constant.stringToRelTime(arg.stringValue());
        } else {
            try {
                return Constant.getInstance(
                    (long) Math.round(toReal(arg) * 1000));
            } catch (NumberFormatException e) {
                return Constant.error(
                            "relTime(invalid number " + arg + ")");
            } catch (ClassCastException e) {
                return Constant.error("relTime(" + arg.typeName() + ")");
            }
        }
    } // relTime(Expr)

    /** Convert to absolute time.
     * @param args the arguments.  May be either a single string (which is
     *        parsed as an ISO 8601 date spec), or an optional number of seconds
     *        offset from the epoch and optional number of seconds east of
     *        Greenwhich.
     * @return an absolute time or error constant.
     */
    public static Expr absTime(Expr[] args) {
        // If the first arg is a string, it must be the only argument.
        // The call is absTime("yyyy-mm-ddThh:mm:ss.mmm+zz:zz") or some such.
        if (args.length > 0 && args[0].type == Expr.STRING) {
            if (args.length > 1) {
                return Constant.error("wrong number of args to absTime");
            }
            return Constant.stringToAbsTime(args[0].stringValue());
        }
        if (args.length > 2) {
            return Constant.error("wrong number of args to absTime");
        }

        long t = 0;
        int z = 0;
        if (args.length == 0) {
            t = System.currentTimeMillis();
            z = TimeZone.getDefault().getOffset(t) / 1000;
        } else {
            try {
                t = Math.round(toReal(args[0]) * 1000);
            } catch (NumberFormatException e) {
                return Constant.error(
                            "absTime(invalid number " + args[0] + ")");
            } catch (ClassCastException e) {
                return Constant.error("absTime(" + args[0].typeName() + ")");
            }
            if (args.length == 1) {
                z = TimeZone.getDefault().getOffset(t) / 1000;
            } else {
                if (args[1].type != Expr.INTEGER) {
                    return Constant.error(
                        "invalid " + args[1].typeName()
                        + " second argument to absTime");
                }
                z = args[1].intValue();
            }
        }
        return Constant.getInstance(t,z);
    } // absTime(Expr)

    //============= numeric operations =======================================

    /** Returns an integer constant corresponding to the floor of the value
     * of arg.
     * @param arg the input value.
     * @return the floor of arg.
     */
    public static Expr floor(Expr arg) {
        if (arg.type == Expr.INTEGER) {
            return arg;
        }
        try {
            return Constant.getInstance((int)Math.floor(toReal(arg)));
        } catch (NumberFormatException e) {
            return Constant.error("floor(invalid number " + arg + ")");
        } catch (ClassCastException e) {
            return Constant.error("floor(" + arg.typeName() + ")");
        }
    } // floor(Expr)

    /** Returns an integer constant corresponding to the ceiling of the value
     * of arg.
     * @param arg the input value.
     * @return the ceiling of arg.
     */
    public static Expr ceiling(Expr arg) {
        if (arg.type == Expr.INTEGER) {
            return arg;
        }
        try {
            return Constant.getInstance((int)Math.ceil(toReal(arg)));
        } catch (NumberFormatException e) {
            return Constant.error("ceiling(invalid number " + arg + ")");
        } catch (ClassCastException e) {
            return Constant.error("ceiling(" + arg.typeName() + ")");
        }
    } // ceiling(Expr)

    /** Returns an integer constant corresponding to the value of arg rounded
     * to the nearest integer.
     * @param arg the input value.
     * @return the rounded arg.
     */
    public static Expr round(Expr arg) {
        if (arg.type == Expr.INTEGER) {
            return arg;
        }
        try {
            return Constant.getInstance((int)Math.round(toReal(arg)));
        } catch (NumberFormatException e) {
            return Constant.error("round(invalid number " + arg + ")");
        } catch (ClassCastException e) {
            return Constant.error("round(" + arg.typeName() + ")");
        }
    } // round(Expr)

    //============= time and date functions ==================================
    //============= see also relTime and absTime =============================

    /** Get the current time in seconds since the epoch.
     * Obsolescent.
     * @return the current time in seconds.
     */
    public static Expr unixTime() {
        return Constant.getInstance((int)(System.currentTimeMillis() / 1000));
    } // unixTime()

    /** Convert a number of seconds into a time interval string of the form
     * "days+hh:mm:ss".  Obsolescent.
     * @param secs the time interval in seconds.
     * @return a string version of the interval.
     */
    public static Expr timeInterval(Expr secs) {
        if (secs.type != Expr.INTEGER) {
            return Constant.error("timeInterval(" + secs.typeName() + ")");
        }
        int time = secs.intValue();
        int[] base = { 60, 60, 24 };
        char[] sep = { ':', ':', '+' };
        String res = ""; 
        for (int i = 0; i < 3; i++) {
            int n = time % base[i];
            time /= base[i];
            res = n + res;
            if (time==0) { 
                break;
            }
            if (n < 10) {
                res = "0" + res;
            }
            res = sep[i] + res;
        }
        if (time > 0) {
            res = time + res;
        }
        return Constant.getInstance(res);
    } // timeInterval(Expr)

    /** Convert a time constant to a string in "Unix" format, using GMT.
     * This function is obsolescent.
     * @param secs the time in seconds from the epoch.
     * @return a string in Unix "ctime" format.
     */
    public static Expr gmtTime(Expr secs) {
        if (secs.type != Expr.INTEGER) {
            return Constant.error("gmtTime(" + secs.typeName() + ")");
        }
        Date when = new Date(1000L * secs.intValue());
        SimpleDateFormat df
            = new SimpleDateFormat("EEE MMM dd HH:mm:ss 'UTC' yyyy");
        df.setTimeZone(TimeZone.getTimeZone("GMT"));
        String s = df.format(when);
        // SimpleDateFormat appears to have no way of specifying that
        // the day of the month is to be formatted with a leading space
        // rather than a leading zero if it is less than 10.
        if (s.charAt(8)=='0') {
            s = s.substring(0, 8) + ' ' + s.substring(9);
        }
        return Constant.getInstance(s);
    } // gmtTime(Expr)

    /** Convert a time constant to a string in "Unix" format, using GMT.
     * This function is obsolescent.
     * @param secs the time in seconds from the epoch.
     * @return a string in Unix "ctime" format.
     */
    public static Expr gmtTimeString(Expr secs) {
        if (secs.type != Expr.INTEGER) {
            return Constant.error("gmtTimeString(" + secs.typeName() + ")");
        }
        Date when = new Date(1000L * secs.intValue());
        SimpleDateFormat df
            = new SimpleDateFormat("EEE MMM dd HH:mm:ss 'UTC' yyyy");
        df.setTimeZone(TimeZone.getTimeZone("GMT"));
        String s = df.format(when);
        // SimpleDateFormat appears to have no way of specifying that
        // the day of the month is to be formatted with a leading space
        // rather than a leading zero if it is less than 10.
        if (s.charAt(8)=='0') {
            s = s.substring(0, 8) + ' ' + s.substring(9);
        }
        return Constant.getInstance(s);
    } // gmtTime(Expr)

    /** Convert a time constant to a string in "Unix" format, using the local
     * time zone.  This function is obsolescent.
     * @param secs the time in seconds from the epoch.
     * @return a string in Unix "ctime" format.
     */
    public static Expr localTimeString(Expr secs) {
        if (secs.type != Expr.INTEGER) {
            return Constant.error("localTimeString(" + secs.typeName() + ")");
        }
        Date when = new Date(1000L * secs.intValue());
        SimpleDateFormat df
            = new SimpleDateFormat("EEE MMM dd HH:mm:ss zzz yyyy");
        String s = df.format(when);
        // SimpleDateFormat appears to have no way of specifying that
        // the day of the month is to be formatted with a leading space
        // rather than a leading zero if it is less than 10.
        if (s.charAt(8)=='0') {
            s = s.substring(0, 8) + ' ' + s.substring(9);
        }
        return Constant.getInstance(s);
    } // localTime(Expr)
        
    //============= private helper methods ===================================

    /** Helper function:  Convert an arbitrary Expr into a String.
     * If the Expr is already a string Constant, just return its value.
     * Otherwise, convert it to a String.
     * @param e the expression to convert.
     * @return the corresponding string.
     */
    private static String getString(Expr e) {
        if (e.type == Expr.STRING) {
            return e.stringValue();
        } else {
            return e.toString();
        }
    } // getString(Expr)

    /** Match pattern against string.
     *  For now, this is pretty simple: "*" matches any sequence of characters,
     *  ? matches any one character, \c matches c, \ at the end matches \ and
     *  any other char matches only itself.
     *  For now, we use a recursive algorithm for simplicity.  This really
     *  ought to be replaced by a more efficient iterative algorithm.
     *  <p><b>TODO:</b>  Add full POSIX glob semantics.
     * @param pat the pattern.
     * @param str the string to be tested for a match.
     * @param ignoreCase if true, lower case in pattern matches upper
     *          case in string (but not vice versa).
     * @return true if the pattern matches the string.
     */
    private static boolean glob(String pat, String str, boolean ignoreCase) {
        if (pat.length() == 0) {
            return str.length()==0;
        }
        char c = pat.charAt(0);
        switch (c) {
        case '?':
            return
                str.length() > 0
                && glob(pat.substring(1),str.substring(1),ignoreCase);
        case '*':
            return
                glob(pat.substring(1),str,ignoreCase)
                || (str.length()>0
                    && glob(pat,str.substring(1),ignoreCase));
        case '\\':
            if (pat.length() == 1)
                return str.equals("\\");
            c = pat.charAt(1);
            return
                str.length()>0
                && str.charAt(0) == c
                && glob(pat.substring(2),str.substring(1),ignoreCase);
        default:
            return
                str.length()>0
                && (str.charAt(0) == c
                    || (ignoreCase
                        && str.charAt(0) == Character.toUpperCase(c)))
                && glob(pat.substring(1),str.substring(1),ignoreCase);
        }
    } // glob(String,String,boolean)

    /** Helper function:  Convert an arbitrary constant into a double value.
     * This is like the method real(), but the result is returned as a double
     * rather than an Expr.
     * @param arg the constant to convert.
     * @exception NumberFormatException
     *      if the constant can not be converted to a real number.
     * @return the corresponding real value
     * @throws ClassCastException if c has the wrong type.
     * @throws NumberFormatException if c cannot be converted to a double.
     */
    private static double toReal(Expr arg) {
        switch (arg.type) {
        case Expr.REAL:
            return arg.realValue();
        case Expr.INTEGER:
            return arg.intValue();
        case Expr.BOOLEAN:
            return arg.isTrue() ? 1.0 : 0.0;
        case Expr.ABSOLUTE_TIME:
            Date d = (Date) ((Constant) arg).value;
            return (double) d.getTime() / 1000;
        case Expr.RELATIVE_TIME:
            Long l = (Long) ((Constant) arg).value;
            return (double) l.longValue() / 1000;
        case Expr.STRING:
            return Constant.stringToDouble(arg.stringValue());
        default:
            throw new ClassCastException("unknown type");
        }
    } // toReal(Constant)

} // Builtin
