/* **************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * Condor Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-2003 Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.
 * Use of the CONDOR Software Program Source Code is authorized
 * solely under the terms of the Condor Public License (see LICENSE.TXT).
 * For more information contact:
 * CONDOR Team, Attention: Professor Miron Livny,
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685,
 * (608) 262-0856 or miron@cs.wisc.edu.
 * ***************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

package condor.classad;

import java.util.*;
import java.text.SimpleDateFormat;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

/**
 * A function call.
 * An internal (non-leaf) node of an expression tree representing a function
 * name and list of arguments.
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.1
 */
public class FuncCall extends Expr {
    private static String VERSION = "$Id: FuncCall.java,v 1.6 2003/10/10 17:13:32 solomon Exp $";

    /** Mapping from names of functions encoded as AttrName objects to
     * FunctionInfo objects.
     */
    private static Map functions = new HashMap();

    /** Information about a function. */
    private static class FunctionInfo {
        /** The method implementing the function. */
        public final Method method;

        /** The number of arguments.  -1 means varags  */
        public final int argc;

        /** Creates a FuncitonInfo object.
         * @param method the method
         * @param argc the argc
         */
        public FunctionInfo(Method method, int argc) {
            this.method = method;
            this.argc = argc;
        } // FunctionInfo(Method, int)

        /** For debugging.
         * @return a string version of this info.
         */
        public String toString() {
            return "(" + method.getName() + "," + argc + ")";
        } // toString()
    } // FunctionInfo

    /** The name of the function as supplied in the original source.  */
    public final AttrName func;

    /** The number of arguments.  -1 means varargs. */
    private int argc;

    /** The method that implements this function.  */
    private Method method;

    /** A pre-allocated error constant to return for unknown functions.
     * Non-null iff method is null.
     */
    private Constant unknownFunction = null;

    /** The parameters to the function. */
    public final Expr[] args;

    /** Pre-load the "built-in" functions. */
    static {
        if (!loadJavaLibrary("condor.classad.Builtin")) {
            throw new RuntimeException(
                        "Cannot load the library condor.classad.Builtin");
        }
    }

    /** Create a function node from a name and a list of paramters.
     * This constructor is private.  Callers should use getInstance instead.
     * @param func the name of the function (original case).
     * @param params the parameters supplied.
     */
    private FuncCall(AttrName func, List params) {
        super(CALL);
        this.func = func;
        if (params == null) {
            this.args = new Expr[0];
        } else {
            this.args = (Expr[]) params.toArray(new Expr[0]);
        }

        // Look up the function in those loaded from libraries.
        FunctionInfo info = (FunctionInfo) functions.get(func);
        if (info != null) {
            this.method = info.method;
            this.argc = info.argc;
            return;
        }

        // In other cases, just return an error on any attempt to evaluate
        // this call.
        unknownFunction = Constant.error("unknown function " + func);
    } // FuncCall(AttrName,List)

    /** "relTime" for constant folding. */
    private static final AttrName REL_TIME = AttrName.fromString("relTime");

    /** "absTime" for constant folding. */
    private static final AttrName ABS_TIME = AttrName.fromString("absTime");

    /** "real" for constant folding. */
    private static final AttrName REAL = AttrName.fromString("real");

    /** Create a function node from a name and a list of paramters.
     * This method has a special hack for "constant folding" of certain
     * calls.  If the function name is absTime or relTime (ignoring
     * case) and there is exactly one argument, of type String, then return the
     * appropriate constant.  In all other cases, just return a new FuncCall
     * object.
     * @param func the name of the function (original case).
     * @param params the parameters supplied.
     * @return an Exp of type CALL, ABS_TIME, REL_TIME, REAL, or ERROR.
     */
    public static Expr getInstance(AttrName func, List params) {
        if (params != null && params.size() == 1) {
            Expr arg = (Expr) params.get(0);
            if (arg.type == STRING) {
                if (func.equals(REL_TIME)) {
                    return Builtin.relTime(arg);
                }
                if (func.equals(ABS_TIME)) {
                    return Constant.stringToAbsTime(arg.stringValue());
                }
            }
        }
        return new FuncCall(func, params);
    } // getInstance(AttrName,List)

    /** Convert this Expr to a string, appending the result to the end of "sb".
     * The representation is the "canonical native format":
     * <code>name '(' [ arg [ ',' arg ]* ')'</code> with no extra spaces.
     * @param sb a place to put the result.
     * @return sb.
     * @see ClassAdWriter
     */
    public StringBuffer toString(StringBuffer sb) {
        sb.append(func).append("(");
        for (int i=0; i<args.length; i++) {
            if (i>0) {
                sb.append(",");
            }
            args[i].toString(sb);
        }
        return sb.append(")");
    } // toString(StringBuffer)

    /** Evaluate this Expr.  This is the internal method used to implement
     * {@link Expr#eval()}.
     * Currently, we assume that all functions are strict in all their
     * arguments, that is, all arguments are evaluated before the function
     * is called, and if any argument evaluates to an undefined or
     * error Constant, the value of the call is undefined or error
     * (with error taking precendence over undefined; the message in a
     * resulting Constant is taken from the first argument that
     * evaluates to an error (or undefined) Constant).
     * <p>We also assume at this time that there are no built-in functions
     * that take a RecordExpr or ListExpr as an argument.
     * @param env an environment used to evaluate the operands.  It is cleared
     * to the null environment before return.
     * @return a Constant representing the value.
     */
    protected Expr eval1(Env env) {
        if (method == null) {
            return unknownFunction;
        }

        int nargs = args.length;
        Expr[] a = new Expr[nargs];
        Expr argval;
        // Evaluate all the args in the original environment
        // This rather complex way of doing it avoids the overhead of
        // cloning the environment if there is exactly one argument.
        for (int i = 0; i < nargs - 1; i++) {
            argval = args[i].eval(new Env(env));
            if (argval.type == ERROR || argval.type == UNDEFINED) {
                return argval;
            }
            a[i] = argval;
        }
        if (nargs > 0) {
            argval = args[nargs - 1].eval(env);
            if (argval.type == ERROR || argval.type == UNDEFINED) {
                return argval;
            }
            a[nargs - 1] = argval;
        }

        try {
            if (argc < 0) {
                return (Expr) method.invoke(null, new Object[] { a });
            } else if (argc != nargs) {
                return Constant.error("wrong number of args to " + func);
            } else {
                return (Expr) method.invoke(null, a);
            }
        } catch (Exception e) {
            return Constant.error(e.getMessage());
        }
    } // eval1(Env)

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * Function calls have maximal precedence (higher than any operator).
     * @return the precendence of this node.
     * @see ClassAdWriter#MINIMAL_PARENTHESES
     */
    public int prec() {
        return MAXPREC+1;
    } // prec()

    /** Loads a library of externally defined Java functions.
     * Meant to be called from ClassAd.loadJavaLibrary.
     * @param className the fully qualified name of the class (e.g.
     *                  condor.classad.Builtin)
     * @return false if the load fails.
     */
    /*package*/ static boolean loadJavaLibrary(String className) {
        try {
            Method[] methods = Class.forName(className).getMethods();
        METHODS:
            for (int i = 0; i < methods.length; i++) {
                Method m = methods[i];
                if (!Modifier.isStatic(m.getModifiers())) {
                    continue;
                }
                if (m.getReturnType() != Expr.class) {
                    continue;
                }
                Class[] param = m.getParameterTypes();
                int argc;
                if (param.length == 1
                        && param[0].getComponentType() == Expr.class)
                {
                    // one param of type Expr[]: varargs
                    argc = -1;
                } else {
                    for (int j = 0; j < param.length; j++) {
                        if (param[j] != Expr.class)  {
                            continue METHODS;
                        }
                    }
                    argc = param.length;
                }
                AttrName name = AttrName.fromString(m.getName());
                if (functions.containsKey(name)) {
                    db("warning: ignoring duplicate function " + name);
                    continue;
                }
                functions.put(name, new FunctionInfo(m, argc));
            }
            return true;
        } catch (ExceptionInInitializerError e) {
            e.printStackTrace(); // debug
            return false;
        } catch (ClassNotFoundException e) {
            e.printStackTrace(); // debug
            return false;
        }
    } // loadJavaLibrary(String)
} // FuncCall
// vim: syntax=java
